/*
 * Copyright 1989 - 1993, Julianne Frances Haugh
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <config.h>

#ident "$Id: faillog.c,v 1.28 2006/06/25 13:08:25 kloczek Exp $"

#include <getopt.h>
#include <pwd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdbool.h>
#include <time.h>
#include "defines.h"
#include "exitcodes.h"
#include "faillog.h"
#include "prototypes.h"
/*
 * Global variables
 */
static FILE *fail;		/* failure file stream */
static uid_t user;		/* one single user, specified on command line */
static int days;		/* number of days to consider for print command */
static time_t seconds;		/* that number of days in seconds */

static bool aflg = false;       /* set if all users are to be printed always */
static bool uflg = false;       /* set if user is a valid user id */
static bool tflg = false;       /* print is restricted to most recent days */
static bool lflg = false;       /* set the locktime */
static bool mflg = false;       /* set maximum failed login counters */
static bool rflg = false;       /* reset the counters of login failures */

static struct stat statbuf;	/* fstat buffer for file size */

#define	NOW	(time((time_t *) 0))

static void usage (void)
{
	fprintf (stderr, _("Usage: faillog [options]\n"
			   "\n"
			   "Options:\n"
			   "  -a, --all			display faillog records for all users\n"
			   "  -h, --help			display this help message and exit\n"
			   "  -l, --lock-time SEC		after failed login lock accout to SEC seconds\n"
			   "  -m, --maximum MAX		set maximum failed login counters to MAX\n"
			   "  -r, --reset			reset the counters of login failures\n"
			   "  -t, --time DAYS		display faillog records more recent than DAYS\n"
			   "  -u, --user LOGIN		display faillog record or maintains failure counters\n"
			   "				and limits (if used with -r, -m or -l options) only\n"
			   "				for user with LOGIN\n"
			   "\n"));
	exit (E_USAGE);
}

static void print_one (const struct passwd *pw, bool force)
{
	static bool once = false;
	struct tm *tm;
	off_t offset;
	struct faillog fl;
	time_t now;

#ifdef HAVE_STRFTIME
	char *cp;
	char ptime[80];
#endif

	if (NULL == pw) {
		return;
	}

	offset = (off_t) pw->pw_uid * sizeof (fl);
	if (offset + sizeof (fl) <= statbuf.st_size) {
		/* fseeko errors are not really relevant for us. */
		fseeko (fail, offset, SEEK_SET);
		/* faillog is a sparse file. Even if no entries were
		 * entered for this user, which should be able to get the
		 * empty entry in this case.
		 */
		if (fread ((char *) &fl, sizeof (fl), 1, fail) != 1) {
			fprintf (stderr,
			         _("faillog: Failed to get the entry for UID %lu\n"),
			         (unsigned long int)pw->pw_uid);
			return;
		}
	} else {
		/* Outsize of the faillog file.
		 * Behave as if there were a missing entry (same behavior
		 * as if we were reading an non existing entry in the
		 * sparse faillog file).
		 */
		memzero (&fl, sizeof (fl));
	}

	/* Nothing to report */
	if (!force && (0 == fl.fail_time)) {
		return;
	}

	(void) time(&now);

	/* Filter out entries that do not match with the -t option */
	if (tflg && ((now - fl.fail_time) > seconds)) {
		return;
	}

	/* Print the header only once */
	if (!once) {
		puts (_("Login       Failures Maximum Latest                   On\n"));
		once = true;
	}

	tm = localtime (&fl.fail_time);
#ifdef HAVE_STRFTIME
	strftime (ptime, sizeof (ptime), "%D %H:%M:%S %z", tm);
	cp = ptime;
#endif
	printf ("%-9s   %5d    %5d   ",
	        pw->pw_name, fl.fail_cnt, fl.fail_max);
	/* FIXME: cp is not defined ifndef HAVE_STRFTIME */
	printf ("%s  %s", cp, fl.fail_line);
	if (0 != fl.fail_locktime) {
		if (   ((fl.fail_time + fl.fail_locktime) > now)
		    && (0 != fl.fail_cnt)) {
			printf (_(" [%lus left]"),
			        (unsigned long) fl.fail_time + fl.fail_locktime - now);
		} else {
			printf (_(" [%lds lock]"),
			        fl.fail_locktime);
		}
	}
	putchar ('\n');
}


static int reset_one (uid_t uid)
{
	off_t offset;
	struct faillog faillog;

	offset = (off_t) uid * sizeof faillog;

	if (offset >= statbuf.st_size)
		return 0;

	if (fseeko (fail, offset, SEEK_SET) != 0) {
		perror (FAILLOG_FILE);
		return 0;
	}
	if (fread ((char *) &faillog, sizeof faillog, 1, fail) != 1) {
		if (!feof (fail))
			perror (FAILLOG_FILE);

		return 0;
	}
	if (faillog.fail_cnt == 0)
		return 1;	/* don't fill in no holes ... */

	faillog.fail_cnt = 0;

	if (fseeko (fail, offset, SEEK_SET) == 0
	    && fwrite ((char *) &faillog, sizeof faillog, 1, fail) == 1) {
		fflush (fail);
		return 1;
	} else {
		perror (FAILLOG_FILE);
	}
	return 0;
}

static void reset (void)
{
	uid_t uid;

	if (uflg)
		reset_one (user);
        else {
                struct passwd *pwent;
 
                setpwent ();
                while ( (pwent = getpwent ()) )
                	reset_one (pwent->pw_uid);
        }
}

static void print (void)
{
	if ( uflg ) {
		print_one (getpwuid (user), true);
	} else {
		/* We only print records for existing users.
		 * Loop based on the user database instead of reading the
		 * whole file. We will have to query the database anyway
		 * so except for very small ranges and large user
		 * database, this should not be a performance issue.
		 */
		struct passwd *pwent;

		setpwent ();
		while ( (pwent = getpwent ()) != NULL ) {
			print_one (pwent, aflg);
		}
		endpwent ();
	}
}



static void setmax_one (uid_t uid, int max)
{
	off_t offset;
	struct faillog faillog;

	offset = (off_t) uid * sizeof faillog;

	if (fseeko (fail, offset, SEEK_SET) != 0) {
		perror (FAILLOG_FILE);
		return;
	}
	if (fread ((char *) &faillog, sizeof faillog, 1, fail) != 1) {
		if (!feof (fail))
			perror (FAILLOG_FILE);
		memzero (&faillog, sizeof faillog);
	}
	faillog.fail_max = max;

	if (fseeko (fail, offset, SEEK_SET) == 0
	    && fwrite ((char *) &faillog, sizeof faillog, 1, fail) == 1)
		fflush (fail);
	else
		perror (FAILLOG_FILE);
}

static void setmax (int max)
{
	struct passwd *pwent;

	if (uflg) {
		setmax_one (user, max);
	} else {
		uid_t uid = 0;
                uid_t uidmax = statbuf.st_size / sizeof (struct faillog);

		while (uid < uidmax) {
			pwent = getpwuid(uid);
			if(pwent)
	                	setmax_one (pwent->pw_uid, max);
                        uid++;
                }
	}
}

static void set_locktime_one (uid_t uid, long locktime)
{
	off_t offset;
	struct faillog faillog;

	offset = (off_t) uid * sizeof faillog;

	if (fseeko (fail, offset, SEEK_SET) != 0) {
		perror (FAILLOG_FILE);
		return;
	}
	if (fread ((char *) &faillog, sizeof faillog, 1, fail) != 1) {
		if (!feof (fail))
			perror (FAILLOG_FILE);
		memzero (&faillog, sizeof faillog);
	}
	faillog.fail_locktime = locktime;

	if (fseeko (fail, offset, SEEK_SET) == 0
	    && fwrite ((char *) &faillog, sizeof faillog, 1, fail) == 1)
		fflush (fail);
	else
		perror (FAILLOG_FILE);
}

/*
 * XXX - this needs to be written properly some day, right now it is
 * a quick cut-and-paste hack from the above two functions.  --marekm
 */
static void set_locktime (long locktime)
{
	struct passwd *pwent;

	if (uflg) {
		set_locktime_one (user, locktime);
	} else {
		setpwent ();
		while ((pwent = getpwent ()))
			set_locktime_one (pwent->pw_uid, locktime);
	}
}

int main (int argc, char **argv)
{
	long fail_locktime;
	long fail_max;
	struct passwd *pwent;

	setlocale (LC_ALL, "");
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);

	/* try to open for read/write, if that fails - read only */
	fail = fopen (FAILLOG_FILE, "r+");
	if (!fail)
		fail = fopen (FAILLOG_FILE, "r");
	if (!fail) {
		perror (FAILLOG_FILE);
		exit (1);
	}

	/* Get the size of the faillog */
	if (fstat (fileno (fail), &statbuf) != 0) {
		fprintf (stderr,
		         _("faillog: Cannot get the size of %s: %s\n"),
		         FAILLOG_FILE, strerror (errno));
		exit (1);
	}

	{
		int option_index = 0;
		int c;
		static struct option long_options[] = {
			{"all", no_argument, NULL, 'a'},
			{"help", no_argument, NULL, 'h'},
			{"lock-secs", required_argument, NULL, 'l'},
			{"maximum", required_argument, NULL, 'm'},
			{"reset", no_argument, NULL, 'r'},
			{"time", required_argument, NULL, 't'},
			{"user", required_argument, NULL, 'u'},
			{NULL, 0, NULL, '\0'}
		};

		while ((c =
			getopt_long (argc, argv, "ahl:m:rt:u:",
				     long_options, &option_index)) != -1) {
			switch (c) {
			case 'a':
				aflg++;
				break;
			case 'h':
				usage ();
				break;
			case 'l':
				fail_locktime = (long) atoi (optarg);
				lflg++;
				break;
			case 'm':
				fail_max = (long) atoi (optarg);
				mflg++;
				break;
			case 'r':
				rflg++;
				break;
			case 't':
				days = atoi (optarg);
				seconds = days * DAY;
				tflg++;
				break;
			case 'u':
				pwent = getpwnam (optarg);
				if (!pwent) {
					fprintf (stderr,
						 _("Unknown User: %s\n"),
						 optarg);
					exit (1);
				}
				uflg++;
				user = pwent->pw_uid;
				break;
			default:
				usage ();
			}
		}
	}

	if (aflg && uflg) {
		usage ();
	}
	if (tflg && (lflg || mflg || rflg)) {
		usage ();
	}

	if (lflg) {
		set_locktime (fail_locktime);
	}

	if (mflg) {
		setmax (fail_max);
	}

	if (rflg) {
		reset ();
	}

	if (!(lflg || mflg || rflg)) {
		print ();
	}

	fclose (fail);

	exit (E_SUCCESS);
}
